import sys
import json 
from collections import namedtuple
from io import StringIO
import base64
import binascii
import json
from collections import defaultdict

BARRIER = '# TEST AREA #'
ENCODING = 'utf8'

def load_notebook(filename: str) -> str:
    '''Read Jupyter notebook and return concatenated contents of all code cells. Ignore any lines after the BARRIER.'''
    with open(filename, encoding=ENCODING) as r:
        ntb = json.load(r)
    cells = [cell['source'] for cell in ntb['cells'] if cell['cell_type']=='code']
    all_lines = []
    for cell in cells:
        for line in cell:
            all_lines.append(line.rstrip())
    code = '\n'.join(all_lines)
    if BARRIER not in code:
        raise EOFError(f'Tested notebook must contain a test area starting with "{BARRIER}"')
    code = code.split(BARRIER, maxsplit=1)[0]
    return code

class Redirecting:
    def __init__(self, stdin=''):
        self.new_in = StringIO(stdin)
        self.new_out = StringIO()
        self.new_err = StringIO()
    def __enter__(self):
        self.old_in = sys.stdin
        self.old_out = sys.stdout
        self.old_err = sys.stderr
        sys.stdin = self.new_in
        sys.stdout = self.new_out
        sys.stderr = self.new_err
        return self
    def __exit__(self, ex_type, ex_value, ex_traceback):
        sys.stdin = self.old_in
        sys.stdout = self.old_out
        sys.stderr = self.old_err
        self.stdout = self.new_out.getvalue()
        self.stderr = self.new_err.getvalue()
        self.new_in.close()
        self.new_out.close()
        self.new_err.close()
        self.exception = ex_value
        self.traceback = ex_traceback
        return True
    def __repr__(self):
        return f'Redirecting(stdout={repr(self.stdout)}, stderr={repr(self.stderr)}, exception={repr(self.exception)}, traceback={repr(self.traceback)})'

def exec_redirected(code: str, *, stdin='', local_vars=None) -> Redirecting:
    '''Execute code, with stdin as the code's standard input, return Redirecting object with the captured outputs.'''
    if local_vars is None:
        local_vars = {}
    with Redirecting(stdin=stdin) as redir:
        exec(code, {}, local_vars)
    return redir

TestCase = namedtuple('TestCase', ['suite_name', 'case_name', 'type', 'visibility', 'input', 'output'])
TestResult = namedtuple('TestResult', ['ok', 'message'])

def test_code_io(case: TestCase, code: str) -> TestResult:
    '''Execute code as compare it's standard output with expected_output'''
    result = exec_redirected(code, stdin=case.input)
    if result.exception is not None:
        ok = False
        lines = [
            f'Test {case.case_name} failed - exception occurred', 
            f'{result.exception.__class__.__name__}: {result.exception}']
    elif result.stdout.rstrip() == case.output.rstrip():
        ok = True
        lines = [f'Test {case.case_name} passed']
    else:
        ok = False
        lines = [f'Test {case.case_name} failed - wrong answer']
        if case.visibility == 'open':
            lines.extend([
                '', 'Test input:', case.input, 
                '', 'Correct output:', case.output, 
                '', 'Your output:', result.stdout
            ])
        else:
            lines.extend(['(details hidden)'])
    return TestResult(ok=ok, message='\n'.join(lines))

class InvalidTestFileError(BaseException):
    pass

class TestSuite(object):
    def __repr__(self) -> str:
        return ''
    def __eq__(self, other) -> bool:
        return (self.__class__ == other.__class__
            and self.name == other.name
            and self.type == other.type 
            and self.cases == other.cases)

class TestSuiteIO(TestSuite):
    def __init__(self, suite_name):
        self.name = suite_name
        self.type = 'io'
        self.cases = []
    def add_case(self, visibility, input, output, case_name=None):
        assert isinstance(input, str)
        assert isinstance(output, str)
        if case_name is None:
            i = len(self.cases) + 1
            case_name = f'#{i}'
        self.cases.append(TestCase(self.name, case_name, 'io', visibility, input, output))
    
def run_test_suite(suite: TestSuite, code: str) -> TestResult:
    for case in suite.cases:
        result = run_test_case(case, code)
        if not result.ok:
            return result
    return TestResult(True, 'Tests passed :)')

def run_test_case(case: TestCase, code: str) -> TestResult:
    if case.type == 'io':
        return test_code_io(case, code)
    else:
        raise NotImplementedError()

def encode_obj(obj: object) -> bytes:
    string = json.dumps(obj)
    check = json.loads(string)
    if check != obj:
        raise ValueError(f'Could not encode object: {obj}')
    return base64.encodebytes(string.encode(ENCODING))

def decode_obj(enc_obj: bytes) -> object:
    string = base64.decodebytes(enc_obj).decode(ENCODING)
    return json.loads(string)

def encode_test_case(case: TestCase) -> bytes:
    result = b' '.join(encode_obj(f) for f in case)
    return result

def decode_test_case(enc_case: bytes) -> TestCase:
    fields = [decode_obj(f) for f in enc_case.split(b' ')]
    return TestCase(*fields)

def simple_hash(message: bytes) -> bytes:
    n = sum(message)
    return base64.encodebytes(str(n).encode(ENCODING))

def check_simple_hash(message: bytes, hash: bytes) -> bool:
    return simple_hash(message) == hash

def encode_tests(test_suites):
    enc_cases = []
    suite_set = set()
    for suite in test_suites:
        if suite.name in suite_set:
            raise ValueError(f'Suite names are not unique: {suite.name}')
        else:
            suite_set.add(suite.name)
        for case in suite.cases:
            enc_cases.append(encode_test_case(case))
    together = b' '.join(base64.encodebytes(c) for c in enc_cases)
    return base64.encodebytes(simple_hash(together) + b' ' + together)
    
def decode_tests(tests):
    tests = base64.decodebytes(tests)
    if b' ' not in tests:
        raise InvalidTestFileError('Invalid test file')
    h, together = tests.split(b' ', 1)
    if not check_simple_hash(together, h):
        raise InvalidTestFileError('Invalid test file')
    enc_cases = [base64.decodebytes(c) for c in together.split(b' ')]
    cases = [decode_test_case(c) for c in enc_cases]
    suites = {}
    for case in cases:
        suite_name = case.suite_name
        if suite_name not in suites:
            suite_type = case.type
            if suite_type == 'io':
                suite = TestSuiteIO(suite_name)
            else:
                raise NotImplementedError
            suites[suite_name] = suite
        suites[suite_name].cases.append(case)
    return list(suites.values())

def save_tests(test_suites, filename):
    content = encode_tests(test_suites)
    with open(filename, 'wb') as w:
        w.write(content)
    check = load_tests(filename)
    if check != test_suites:
        raise ValueError(f'Could not encode tests: {test_suites}')

def load_tests(filename):
    try:
        with open(filename, 'rb') as r:
            content = r.read()
        tests = decode_tests(content)
        return tests
    except OSError:
        raise InvalidTestFileError(f'Cannot read test file "{filename}". Make sure you did not remove or rename this file.')
    except binascii.Error:
        raise InvalidTestFileError(f'Invalid test file "{filename}". Make sure you did not modify this file.')
    
def load_suite(test_file: str, suite_name: str) -> TestSuite:
    tests = load_tests(test_file)
    try:
        return next(s for s in tests if s.name == suite_name)
    except StopIteration:
        raise InvalidTestFileError(f'Test suite {suite_name} not found')

def run_tests(test_file: str, test_suite: str, notebook: str):
    '''Run all tests in suite test_suite in test_file on the merged code cells in notebook.'''
    try:
        code = load_notebook(notebook)
    except OSError:
        print('ERROR:', f'Cannot read file "{notebook}". Make sure you did not rename the notebook.', end='\n\n', file=sys.stderr)
        return
    except EOFError as ex:
        print('ERROR:', ex, end='\n\n', file=sys.stderr)
        return
    try:
        suite = load_suite(test_file, test_suite)
    except InvalidTestFileError as ex:
        print('ERROR:', ex, end='\n\n', file=sys.stderr)
        return
    result = run_test_suite(suite, code)
    print(result.message, file=sys.stderr)
